/*********************************************************************
* Software License Agreement (BSD License)
*
*  Copyright (c) 2010, Rice University
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
*
*   * Redistributions of source code must retain the above copyright
*     notice, this list of conditions and the following disclaimer.
*   * Redistributions in binary form must reproduce the above
*     copyright notice, this list of conditions and the following
*     disclaimer in the documentation and/or other materials provided
*     with the distribution.
*   * Neither the name of the Rice University nor the names of its
*     contributors may be used to endorse or promote products derived
*     from this software without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
*  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
*  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
*  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
*  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
*  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*  POSSIBILITY OF SUCH DAMAGE.
*********************************************************************/

/* Author: Mark Moll */

#ifndef OMPL_EXTENSION_VORTEX_ENVIRONMENT_
#define OMPL_EXTENSION_VORTEX_ENVIRONMENT_

#include "ompl/config.h"
#if OMPL_EXTENSION_VORTEX == 0
#  error Vortex extension not built
#endif

#include "ompl/util/ClassForward.h"
#include "ompl/util/Exception.h"
#include "ompl/base/State.h"

#include <Vx/VxFrame.h>
#include <Vx/VxUniverse.h>
#include <Vx/VxIntersectResult.h>
#include <VxGraphics/VxVisualizer.h>

#include <vector>
#include <boost/thread/mutex.hpp>

namespace ompl
{
    namespace control
    {

        /** \brief Forward declaration of ompl::control::VortexEnvironment */
        ClassForward(VortexEnvironment);

        /** \class ompl::control::VortexEnvironmentPtr
            \brief A boost shared pointer wrapper for ompl::control::VortexEnvironment */

        /** \brief This class contains the Vortex constructs OMPL needs to know about when planning. */
        class VortexEnvironment
        {
        public:
            /** \brief Subscriber class to be notified of creation/deletion of contacts.

                It is also possible to change the collision response, e.g., by adding
                forces to the colliding bodies. */
            class VortexCollisionSubscriber : public Vx::VxUniverse::IntersectSubscriber
            {
            public:
                VortexCollisionSubscriber(VortexEnvironment* env)
                    : env_(env) {}

                /** \brief Callback for the creation of contacts */
                virtual void notifyIntersect(Vx::VxUniverse::eIntersectEventType type,
                    Vx::VxIntersectResult *result, Vx::VxDynamicsResponseInput* response);

                /** \brief Callback for breaking contact */
                virtual void notifyDisjoint(Vx::VxUniverse::eIntersectEventType type,
                    Vx::VxIntersectResult *result) {}

                VortexEnvironment* env_;
            };

            /** \brief The Vortex world where the simulation is performed */
            Vx::VxFrame*             world_;

            /** \brief The set of objects that need to be considered
                part of the state when planning. These are not
                necessarily all the objects in the environment.*/
            Vx::VxAssemblyBase*      assembly_;

            /** \brief Issue debug messages when contacts are found. Default is false.
                This should only be used for debugging */
            bool                     verboseContacts_;

            /** \brief The simulation step size */
            double                   stepSize_;

            /** \brief The maximum number of times a control is applied in sequence */
            unsigned int             maxControlSteps_;

            /** \brief The minimum number of times a control is applied in sequence */
            unsigned int             minControlSteps_;

            /** Collision subscriber used whenever a state or path is checked for
                valid collisions */
            VortexCollisionSubscriber* collisionSubscriber_;

            /** Flag that indicates whether an invalid collision occured during the
                last state check/propagation */
            bool                     invalidCollisionOccured_;

            /** Vortex visualizer, used during playback of a path */
            VxGraphics::VxVisualizer* visualizer_;

            /** If non-zero, pointer to a VxEntity (such as a VxPart) that the camera will follow */
            Vx::VxEntity*            followEntity_;

            /** \brief Lock to use when performing simulations in the world, since
                the environment is shared among all threads */
            mutable boost::mutex  mutex_;

            VortexEnvironment(void)
                : world_(Vx::VxFrame::instance()), verboseContacts_(false), stepSize_(0.05),
                maxControlSteps_(1000), minControlSteps_(5), collisionSubscriber_(NULL),
                invalidCollisionOccured_(false), visualizer_(VxGraphics::VxVisualizer::create()),
                followEntity_(NULL)
            {
                Vx::VxFrameRegisterAllInteractions(world_);
            }

            virtual ~VortexEnvironment(void)
            {
                world_->release();
                delete visualizer_;
                if (collisionSubscriber_)
                    delete collisionSubscriber_;
            }

            /** \brief Cast this instance to a desired type. */
            template<class T>
            T* as(void)
            {
                /** \brief Make sure the type we are casting to is indeed a Vortex environment */
                BOOST_CONCEPT_ASSERT((boost::Convertible<T*, VortexEnvironment*>));
                return static_cast<T*>(this);
            }

            /** \brief Number of parameters (double values) needed to specify a control input */
            virtual unsigned int getControlDimension(void) const = 0;

            /** \brief Get the control bounds -- the bounding box in which to sample controls */
            virtual void getControlBounds(std::vector<double> &lower, std::vector<double> &upper) const = 0;

            /** \brief Application of a control. This function sets
                the forces/torques/velocities for bodies in the
                simulation based on control inputs.*/
            virtual void applyControl(const double *control) const = 0;

            /** \brief Decide whether a collision is a valid one or
                not. In some cases, collisions between some geometries can
                be allowed. By default, this function always returns
                false, making all collisions invalid */
            virtual bool isValidCollision(const Vx::VxIntersectResult& contact) const
            {
                return false;
            }

            /** \brief Add a collision subscriber. This is subscriber will be used in
                \b addition to the default one.

                The parameters type1 and type2 specify the types of objects that are colliding,
                while eventID specifies whether this is new collision, ongoing collision, etc.
                See the Vortex documentation and header files for more information. */
            virtual void addCollisionSubscriber(VortexCollisionSubscriber* subscriber,
                Vx::VxUniverse::ResponseType type1 = Vx::VxUniverse::kResponsePart,
                Vx::VxUniverse::ResponseType type2 = Vx::VxUniverse::kResponsePart,
                Vx::VxUniverse::eIntersectEventType eventID = Vx::VxUniverse::kEventFirst);

            virtual void addDefaultCollisionSubscriber();
        };


    }
}

#endif
