/*********************************************************************
* Software License Agreement (BSD License)
*
*  Copyright (c) 2010, Rice University
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
*
*   * Redistributions of source code must retain the above copyright
*     notice, this list of conditions and the following disclaimer.
*   * Redistributions in binary form must reproduce the above
*     copyright notice, this list of conditions and the following
*     disclaimer in the documentation and/or other materials provided
*     with the distribution.
*   * Neither the name of the Rice University nor the names of its
*     contributors may be used to endorse or promote products derived
*     from this software without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
*  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
*  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
*  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
*  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
*  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*  POSSIBILITY OF SUCH DAMAGE.
*********************************************************************/

/* Author: Mark Moll */

#include "ompl/extensions/vortex/VortexSimpleSetup.h"
#include "ompl/extensions/vortex/VortexStatePropagator.h"
#include "ompl/util/Exception.h"
#include <Vx/VxEntity.h>
#include <boost/thread.hpp>

ompl::control::VortexSimpleSetup::VortexSimpleSetup(const ControlSpacePtr &space) : SimpleSetup(space)
{
    if (!dynamic_cast<VortexControlSpace*>(space.get()))
        throw Exception("Vortex Control Space needed for Vortex Simple Setup");
    useEnvParams();
}

ompl::control::VortexSimpleSetup::VortexSimpleSetup(const base::StateSpacePtr &space) :
    SimpleSetup(ControlSpacePtr(new VortexControlSpace(space)))
{
    useEnvParams();
}

ompl::control::VortexSimpleSetup::VortexSimpleSetup(const VortexEnvironmentPtr &env) :
    SimpleSetup(ControlSpacePtr(new VortexControlSpace(base::StateSpacePtr(new VortexStateSpace(env)))))
{
    useEnvParams();
}

void ompl::control::VortexSimpleSetup::useEnvParams(void)
{
    const VortexEnvironmentPtr& env
        = getStateSpace()->as<VortexStateSpace>()->getEnvironment();
    si_->setPropagationStepSize(10*env->stepSize_);
    si_->setMinMaxControlDuration(env->minControlSteps_, env->maxControlSteps_);
    si_->setStatePropagator(StatePropagatorPtr(new VortexStatePropagator(si_)));
}

ompl::base::ScopedState<ompl::control::VortexStateSpace> ompl::control::VortexSimpleSetup::getCurrentState(void) const
{
    base::ScopedState<VortexStateSpace> current(getStateSpace());
    getStateSpace()->as<VortexStateSpace>()->readState(current.get());
    return current;
}

void ompl::control::VortexSimpleSetup::setCurrentState(const base::State *state)
{
    getStateSpace()->as<VortexStateSpace>()->writeState(state);
}

void ompl::control::VortexSimpleSetup::setCurrentState(const base::ScopedState<> &state)
{
    getStateSpace()->as<VortexStateSpace>()->writeState(state.get());
}

void ompl::control::VortexSimpleSetup::setup(void)
{
    if (!si_->getStateValidityChecker())
    {
        msg_.inform("Using default state validity checker for Vortex");
        si_->setStateValidityChecker(base::StateValidityCheckerPtr(new VortexStateValidityChecker(si_)));
    }
    if (pdef_->getStartStateCount() == 0)
    {
        msg_.inform("Using the initial state of Vortex as the starting state for the planner");
        pdef_->addStartState(getCurrentState());
    }
    SimpleSetup::setup();
}

bool ompl::control::VortexSimpleSetup::playSolutionPath(double timeFactor) const
{
    return (haveSolutionPath())
        ? playPath(getGoal()->getSolutionPath(), timeFactor)
        : false;
}

bool ompl::control::VortexSimpleSetup::playPath(const base::PathPtr &path, double timeFactor) const
{
    bool ctl = false, running = false;
    const VortexEnvironmentPtr& env
        = getStateSpace()->as<VortexStateSpace>()->getEnvironment();

    if (dynamic_cast<PathControl*>(path.get()))
        ctl = true;
    else
        if (!dynamic_cast<geometric::PathGeometric*>(path.get()))
            throw Exception("Unknown type of path");

    geometric::PathGeometric pg(ctl ?
        static_cast<PathControl*>(path.get())->asGeometric() : *static_cast<geometric::PathGeometric*>(path.get()));
    double duration = timeFactor * si_->getPropagationStepSize() * (double)(pg.states.size() - 1);
    double timeStep = 1./20.;

    if (!pg.states.empty())
    {
        pg.interpolate(ceil(duration/timeStep));

        time::duration d = time::seconds(timeStep);
        getStateSpace()->as<VortexStateSpace>()->writeState(pg.states[0]);
        env->world_->preStepUpdate();
        env->world_->updateCollision();
        env->world_->postStepUpdate();
        running = env->visualizer_->update();

        for (unsigned int i = 1 ; i < pg.states.size() ; )
        {
            if (!env->visualizer_->isPaused())
            {
                boost::this_thread::sleep(d);
                getStateSpace()->as<VortexStateSpace>()->writeState(pg.states[i++]);
                env->world_->preStepUpdate();
                env->world_->updateCollision();
                env->world_->postStepUpdate();
            }
            if (env->followEntity_)
                env->visualizer_->setCameraLookAt(env->followEntity_->getPosition());
            running = env->visualizer_->update();
        }
    }
    return running;
}

ompl::base::PathPtr ompl::control::VortexSimpleSetup::simulateControl(const double* control, double duration) const
{
    Control *c = si_->allocControl();
    memcpy(c->as<VortexControlSpace::ControlType>()->values, control, sizeof(double) * getControlSpace()->getDimension());
    base::PathPtr path = simulateControl(c, duration);
    si_->freeControl(c);
    return path;
}

ompl::base::PathPtr ompl::control::VortexSimpleSetup::simulateControl(const Control* control, double duration) const
{
    PathControl *p(new PathControl(si_));

    base::State *s0 = si_->allocState();
    getStateSpace()->as<VortexStateSpace>()->readState(s0);
    p->states.push_back(s0);

    base::State *s1 = si_->allocState();
    si_->propagate(s0, control, duration, s1);
    p->states.push_back(s1);

    p->controls.push_back(si_->cloneControl(control));
    p->controlDurations.push_back(duration);
    return base::PathPtr(p);
}

ompl::base::PathPtr ompl::control::VortexSimpleSetup::simulate(double duration) const
{
    Control *c = si_->allocControl();
    si_->nullControl(c);
    base::PathPtr path = simulateControl(c, duration);
    si_->freeControl(c);
    return path;
}
